/**
 * Copyright (c) 2014-2016 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see http://www.gnu.org/licenses/
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.registry_overlay.core;

import static org.junit.Assert.assertEquals;

import java.io.File;
import java.net.URISyntaxException;
import java.net.URL;
import java.util.Properties;
import java.util.concurrent.Callable;

import org.hamcrest.Matchers;
import org.junit.AfterClass;
import org.junit.Assert;
import org.junit.BeforeClass;
import org.junit.Test;
import org.ow2.petals.registry_overlay.api.Constants;
import org.ow2.petals.registry_overlay.api.DataModelApi;
import org.ow2.petals.registry_overlay.api.StoredRegistryEndpoint;
import org.ow2.petals.registry_overlay.api.StoredRegistryEndpoint.Type;
import org.ow2.petals.registry_overlay.configuration.generated.PetalsRegistryOverlay;
import org.ow2.petals.registry_overlay.core.exception.ConfigurationException;
import org.ow2.petals.topology.generated.Container;
import org.ow2.petals.topology.generated.Domain;
import org.ow2.petals.topology.generated.Topology;

import com.ebmwebsourcing.easycommons.lang.reflect.ReflectionHelper;
import com.hazelcast.client.HazelcastClient;
import com.hazelcast.core.HazelcastInstance;
import com.hazelcast.core.IMap;
import com.jayway.awaitility.Awaitility;
import com.jayway.awaitility.Duration;

/**
 * Unit tests for topology cleaner listener
 * 
 * @author Christophe DENEUX - Linagora
 * 
 */
public class TopologyCleanerListenerTest {

    private static RegistryOverlayNode NODE;

    private static RegistryOverlayNode NODE2;

    @BeforeClass
    public static void setupLog() throws URISyntaxException {
        final URL jdkLogCfgFileUrl = Thread.currentThread().getContextClassLoader().getResource("logging.properties");
        System.setProperty("java.util.logging.config.file", new File(jdkLogCfgFileUrl.toURI()).getAbsolutePath());
    }

    @BeforeClass
    public static void startHzlCluster() throws ConfigurationException {
        final PetalsRegistryOverlay petalsRegistryOverlay = ClusterBuilder
                .createVeryBasicClusterConfig(Constants.DEFAULT_PORT, Constants.DEFAULT_PORT + 1);
        final Properties memberProperties = new Properties();
        memberProperties.setProperty(MemberConstants.MEMBER_PROPERTY_NAME_MEMBER_ID, ClusterBuilder.MEMBER_ID_1);
        NODE = new RegistryOverlayNode(petalsRegistryOverlay, memberProperties);
        NODE.start();

        final Properties memberProperties2 = new Properties();
        memberProperties2.setProperty(MemberConstants.MEMBER_PROPERTY_NAME_MEMBER_ID, ClusterBuilder.MEMBER_ID_2);
        NODE2 = new RegistryOverlayNode(petalsRegistryOverlay, memberProperties2);
        NODE2.start();
    }

    private static HazelcastInstance getInstance() {
        return getInstance(NODE);
    }

    @SuppressWarnings("deprecation")
    private static HazelcastInstance getInstance(final RegistryOverlayNode node) {
        return (HazelcastInstance) ReflectionHelper.getPrivateFieldValue(RegistryOverlayNode.class, node, "instance");
    }

    @AfterClass
    public static void shutdownHzlCluster() {
        NODE.stop();
        NODE2.stop();
    }

    /**
     * <p>
     * Check the processing of the event on a registry containing two different standalone topologies.
     * </p>
     * <p>
     * Expected result:
     * <ul>
     * <li>no error occur,</li>
     * <li>on a first container remove, the information associated to topology of the container are cleaned,</li>
     * <li>on a 2nd container remove, all information are cleaned,</li>
     * </ul>
     * </p>
     */
    @Test
    public void oneMemberContainingAnotherTopology() throws InterruptedException {

        final HazelcastInstance client1 = HazelcastClient.newHazelcastClient(ClusterBuilder
                .createClientConfigForVeryBasicCluster());

        final HazelcastInstance client2 = HazelcastClient
                .newHazelcastClient(ClusterBuilder.createClientConfigForVeryBasicCluster());

        final HazelcastInstance instance = getInstance();

        assertEquals(2, instance.getClientService().getConnectedClients().size());

        final String domain_1 = "domain_1";
        final String interfaceName_1 = "interface-name-1";
        final String serviceName_1 = "service-name-1";
        final String endpointName_1 = "endpoint-name-1";

        final String domain_2 = "domain_2";
        final String interfaceName_2 = "interface-name-2";
        final String serviceName_2 = "service-name-2";
        final String endpointName_2 = "endpoint-name-2";

        try {

            final StoredRegistryEndpoint endpoint_1 = new StoredRegistryEndpoint(Type.INTERNAL, endpointName_1,
                    serviceName_1, new String[] { interfaceName_1 }, "", "");

            {
                final Topology topology = new Topology();
                topology.setDomain(new Domain());
                topology.getDomain().getContainer().add(new Container());
                final Topology prev = DataModelApi.getTopologiesMap(client1).put(domain_1, topology);
                Assert.assertNull(prev);
            }

            DataModelApi.getEndpointsMap(client1, domain_1).set(endpointName_1, endpoint_1);
            DataModelApi.getEndpointDescriptionsMap(client1, domain_1).set(endpointName_1, "endpoint-description");

            final StoredRegistryEndpoint endpoint_2 = new StoredRegistryEndpoint(Type.INTERNAL, endpointName_2,
                    serviceName_2, new String[] { interfaceName_2 }, "", "");

            {
                final Topology topology = new Topology();
                topology.setDomain(new Domain());
                topology.getDomain().getContainer().add(new Container());
                final Topology prev = DataModelApi.getTopologiesMap(client2).put(domain_2, topology);
                Assert.assertNull(prev);
            }

            DataModelApi.getEndpointsMap(client2, domain_2).set(endpointName_2, endpoint_2);
            DataModelApi.getEndpointDescriptionsMap(client2, domain_2).set(endpointName_2, "endpoint-description");

            assertEquals(1, DataModelApi.getEndpointDescriptionsMap(instance, domain_1).size());
            assertEquals(1, DataModelApi.getEndpointsMap(instance, domain_1).size());
            assertEquals(1, DataModelApi.getEndpointDescriptionsMap(instance, domain_2).size());
            assertEquals(1, DataModelApi.getEndpointsMap(instance, domain_2).size());
            assertEquals(2, DataModelApi.getTopologiesMap(instance).size());

            // now let's update the second domain to simulate that a member left
            {
                final Topology topology = DataModelApi.getTopologiesMap(client2).get(domain_2);
                topology.getDomain().getContainer().remove(0);
                DataModelApi.getTopologiesMap(client2).set(domain_2, topology);
            }

            // We wait that the event is processed
            Awaitility.await().atMost(Duration.FIVE_SECONDS).until(mapSize(DataModelApi.getTopologiesMap(instance)),
                    Matchers.equalTo(1));

            assertEquals(0, DataModelApi.getEndpointDescriptionsMap(instance, domain_2).size());
            assertEquals(0, DataModelApi.getEndpointsMap(instance, domain_2).size());

            assertEquals(1, DataModelApi.getEndpointDescriptionsMap(instance, domain_1).size());
            assertEquals(1, DataModelApi.getEndpointsMap(instance, domain_1).size());

            // now let's update the first domain to simulate that a member left
            {
                final Topology topology = DataModelApi.getTopologiesMap(client1).get(domain_1);
                topology.getDomain().getContainer().remove(0);
                DataModelApi.getTopologiesMap(client1).set(domain_1, topology);
            }

            // We wait that the event is processed
            Awaitility.await().atMost(Duration.FIVE_SECONDS).until(mapSize(DataModelApi.getTopologiesMap(instance)),
                    Matchers.equalTo(0));

            assertEquals(0, DataModelApi.getEndpointDescriptionsMap(instance, domain_1).size());
            assertEquals(0, DataModelApi.getEndpointsMap(instance, domain_1).size());

        } finally {
            client1.shutdown();
            client2.shutdown();
        }
    }

    public Callable<Integer> mapSize(final IMap<?, ?> map) {
        return new Callable<Integer>() {
            @Override
            public Integer call() throws Exception {
                return map.size();
            }
        };
    }

    /**
     * <p>
     * Check the processing of the event on a registry containing one topology composed of two containers.
     * </p>
     * <p>
     * Expected result:
     * <ul>
     * <li>no error occur,</li>
     * <li>on a first container remove, no information is cleaned,</li>
     * <li>on a 2nd container remove, all information are cleaned,</li>
     * </ul>
     * </p>
     */
    @Test
    public void oneMemberContainingOneTopology() throws InterruptedException {

        final HazelcastInstance client1 = HazelcastClient.newHazelcastClient(ClusterBuilder
                .createClientConfigForVeryBasicCluster());

        final HazelcastInstance client2 = HazelcastClient
                .newHazelcastClient(ClusterBuilder.createClientConfigForVeryBasicCluster());

        final HazelcastInstance instance = getInstance();

        assertEquals(2, instance.getClientService().getConnectedClients().size());

        final String domain = "my-domain";
        final String interfaceName_1 = "interface-name-1";
        final String serviceName_1 = "service-name-1";
        final String endpointName_1 = "endpoint-name-1";
        final String interfaceName_2 = "interface-name-2";
        final String serviceName_2 = "service-name-2";
        final String endpointName_2 = "endpoint-name-2";

        final StoredRegistryEndpoint endpoint_1 = new StoredRegistryEndpoint(Type.INTERNAL, endpointName_1,
                serviceName_1, new String[] { interfaceName_1 }, "", "");

        final StoredRegistryEndpoint endpoint_2 = new StoredRegistryEndpoint(Type.INTERNAL, endpointName_2,
                serviceName_2, new String[] { interfaceName_2 }, "", "");

        try {
            {
                final Topology topology = new Topology();
                topology.setDomain(new Domain());
                topology.getDomain().getContainer().add(new Container());
                DataModelApi.getTopologiesMap(client1).set(domain, topology);
            }

            DataModelApi.getEndpointsMap(client1, domain).set(endpointName_1, endpoint_1);
            DataModelApi.getEndpointDescriptionsMap(client1, domain).set(endpointName_1, "endpoint-description");

            {
                final Topology topology = DataModelApi.getTopologiesMap(client2).get(domain);
                topology.getDomain().getContainer().add(new Container());
                DataModelApi.getTopologiesMap(client2).set(domain, topology);
            }

            DataModelApi.getEndpointsMap(client2, domain).set(endpointName_2, endpoint_2);
            DataModelApi.getEndpointDescriptionsMap(client2, domain).set(endpointName_2, "endpoint-description");

            assertEquals(2, DataModelApi.getEndpointDescriptionsMap(instance, domain).size());
            assertEquals(2, DataModelApi.getEndpointsMap(instance, domain).size());
            assertEquals(1, DataModelApi.getTopologiesMap(instance).size());

            // now let's update the first domain to simulate that a member left
            {
                final Topology topology = DataModelApi.getTopologiesMap(client1).get(domain);
                topology.getDomain().getContainer().remove(0);
                DataModelApi.getTopologiesMap(client1).set(domain, topology);
            }

            // We wait that the event is processed
            Thread.sleep(2000);

            // nothing changed
            assertEquals(2, DataModelApi.getEndpointDescriptionsMap(instance, domain).size());
            assertEquals(2, DataModelApi.getEndpointsMap(instance, domain).size());
            assertEquals(1, DataModelApi.getTopologiesMap(instance).size());
            
            // now let's update the second domain to simulate that a member left
            {
                final Topology topology = DataModelApi.getTopologiesMap(client2).get(domain);
                topology.getDomain().getContainer().remove(0);
                DataModelApi.getTopologiesMap(client2).set(domain, topology);
            }

            // We wait that the event is processed
            Awaitility.await().atMost(Duration.FIVE_SECONDS).until(mapSize(DataModelApi.getTopologiesMap(instance)),
                    Matchers.equalTo(0));

            assertEquals(0, DataModelApi.getEndpointDescriptionsMap(instance, domain).size());
            assertEquals(0, DataModelApi.getEndpointsMap(instance, domain).size());
            
        } finally {
            client1.shutdown();
            client2.shutdown();
        }
    }

    @Test
    public void twoMembersContainingOneTopology() throws InterruptedException {
        final HazelcastInstance client1 = HazelcastClient
                .newHazelcastClient(ClusterBuilder.createClientConfigForVeryBasicCluster(Constants.DEFAULT_PORT));

        final HazelcastInstance client2 = HazelcastClient
                .newHazelcastClient(ClusterBuilder.createClientConfigForVeryBasicCluster(Constants.DEFAULT_PORT + 1));

        final HazelcastInstance instance1 = getInstance(NODE);
        final HazelcastInstance instance2 = getInstance(NODE2);

        assertEquals(1, instance1.getClientService().getConnectedClients().size());
        assertEquals(1, instance2.getClientService().getConnectedClients().size());

        final String domain = "my-domain";
        final String interfaceName_1 = "interface-name-1";
        final String serviceName_1 = "service-name-1";
        final String endpointName_1 = "endpoint-name-1";
        final String interfaceName_2 = "interface-name-2";
        final String serviceName_2 = "service-name-2";
        final String endpointName_2 = "endpoint-name-2";

        final StoredRegistryEndpoint endpoint_1 = new StoredRegistryEndpoint(Type.INTERNAL, endpointName_1,
                serviceName_1, new String[] { interfaceName_1 }, "", "");

        final StoredRegistryEndpoint endpoint_2 = new StoredRegistryEndpoint(Type.INTERNAL, endpointName_2,
                serviceName_2, new String[] { interfaceName_2 }, "", "");

        try {

            {
                final Topology topology = new Topology();
                topology.setDomain(new Domain());
                topology.getDomain().getContainer().add(new Container());
                DataModelApi.getTopologiesMap(client1).set(domain, topology);
            }

            DataModelApi.getEndpointsMap(client1, domain).set(endpointName_1, endpoint_1);
            DataModelApi.getEndpointDescriptionsMap(client1, domain).set(endpointName_1, "endpoint-description");

            {
                final Topology topology = DataModelApi.getTopologiesMap(client2).get(domain);
                topology.getDomain().getContainer().add(new Container());
                DataModelApi.getTopologiesMap(client2).set(domain, topology);
            }

            DataModelApi.getEndpointsMap(client2, domain).set(endpointName_2, endpoint_2);
            DataModelApi.getEndpointDescriptionsMap(client2, domain).set(endpointName_2, "endpoint-description");

            assertEquals(2, DataModelApi.getEndpointDescriptionsMap(instance1, domain).size());
            assertEquals(2, DataModelApi.getEndpointsMap(instance1, domain).size());
            assertEquals(1, DataModelApi.getTopologiesMap(instance1).size());

            // now let's update the first domain to simulate that a member left
            {
                final Topology topology = DataModelApi.getTopologiesMap(client1).get(domain);
                topology.getDomain().getContainer().remove(0);
                DataModelApi.getTopologiesMap(client1).set(domain, topology);
            }

            // We wait that the event is processed
            Thread.sleep(2000);

            // nothing changed
            assertEquals(2, DataModelApi.getEndpointDescriptionsMap(instance1, domain).size());
            assertEquals(2, DataModelApi.getEndpointsMap(instance1, domain).size());
            assertEquals(1, DataModelApi.getTopologiesMap(instance1).size());

            // now let's update the second domain to simulate that a member left
            {
                final Topology topology = DataModelApi.getTopologiesMap(client2).get(domain);
                topology.getDomain().getContainer().remove(0);
                DataModelApi.getTopologiesMap(client2).set(domain, topology);
            }

            // We wait that the event is processed
            Awaitility.await().atMost(Duration.FIVE_SECONDS).until(mapSize(DataModelApi.getTopologiesMap(instance1)),
                    Matchers.equalTo(0));

            assertEquals(0, DataModelApi.getEndpointDescriptionsMap(instance1, domain).size());
            assertEquals(0, DataModelApi.getEndpointsMap(instance1, domain).size());

        } finally {
            client1.shutdown();
            client2.shutdown();
        }
    }
}
